import csv
import logging
import multiprocessing
import time
from concurrent.futures import ThreadPoolExecutor
from itertools import repeat

import pandas as pd
import requests

THREAD_POOL = multiprocessing.cpu_count()

session = requests.Session()
session.mount(
    "https://",
    requests.adapters.HTTPAdapter(
        pool_maxsize=THREAD_POOL, max_retries=3, pool_block=True
    ),
)


def get(
    site_name,
    start_date,
    end_date,
    test_run,
    full_csv_name,
    url,
    direction,
    longitude,
    latitude,
    status,
    easting,
    northing,
):

    message = "Parallel request of data for use in ONS. Emerging Platforms Team. @GitHub: dombean/road_data_scraper"

    headers = {"Message": f"{message}"}

    response = session.get(url, headers=headers)

    logging.info(
        f"request was completed in {response.elapsed.total_seconds()} seconds [{site_name}] [{response.url}]"
    )

    dataframe = pd.DataFrame.from_dict(response.json()["Rows"])

    dataframe = dataframe.assign(
        direction=direction,
        longitude=longitude,
        latitude=latitude,
        status=status,
        easting=easting,
        northing=northing,
    )

    dataframe.to_csv(f"{full_csv_name}", mode="a", header=False, index=False)

    if response.status_code != 200:
        logging.error(
            "request failed, error code %s [%s]", response.status_code, response.url
        )

    if 500 <= response.status_code < 600:
        # server is overloaded? give it a break
        time.sleep(5)

    return response


def download(site_name, start_date, end_date, metadata, test_run, run_id_path):

    headers = [
        "site_name",
        "report_date",
        "time_period_ending",
        "time_interval",
        "0_520_cm",
        "521_660_cm",
        "661_1160_cm",
        "1160+_cm",
        "0_10_mph",
        "11_15_mph",
        "16_20_mph",
        "21_25_mph",
        "26_30_mph",
        "31_35_mph",
        "36_40_mph",
        "41_45_mph",
        "46_50_mph",
        "51_55_mph",
        "56_60_mph",
        "61_70_mph",
        "71_80_mph",
        "80_plus_mph",
        "avg_mph",
        "total_volume",
        "direction",
        "longitude",
        "latitude",
        "status",
        "easting",
        "northing",
    ]

    if test_run:
        full_csv_name = (
            f"{str(run_id_path)}/{site_name}_{start_date}-{end_date}_TEST_RUN.csv"
        )
    else:
        full_csv_name = f"{str(run_id_path)}/{site_name}_{start_date}-{end_date}.csv"

    with open(full_csv_name, "w") as csvfile:
        writer = csv.DictWriter(csvfile, fieldnames=headers)
        writer.writeheader()

    url, direction, longitude, latitude, status, easting, northing = zip(*metadata)

    with ThreadPoolExecutor(max_workers=THREAD_POOL) as executor:
        executor.map(
            get,
            repeat(site_name),
            repeat(start_date),
            repeat(end_date),
            repeat(test_run),
            repeat(full_csv_name),
            url,
            direction,
            longitude,
            latitude,
            status,
            easting,
            northing,
        )



import ast
import calendar
import configparser
import logging
import shutil
import time
from datetime import datetime
from pathlib import Path

from dateutil.relativedelta import relativedelta
from rich.logging import RichHandler

from road_data_scraper.report.report import run_reports
from road_data_scraper.steps.download import THREAD_POOL, download
from road_data_scraper.steps.file_handler import (
    dump_config,
    file_handler,
    gcp_upload_from_directory,
)
from road_data_scraper.steps.metadata import get_site_urls, get_sites_by_sensor

logging.basicConfig(
    format="%(asctime)s.%(msecs)03d %(levelname)-8s %(message)s",
    level=logging.INFO,
    datefmt="%Y-%m-%d %H:%M:%S",
    handlers=[RichHandler()],
)


def run(config, api_run):

    if api_run:

        def my_ast(*args):
            return args[0]

    else:

        def my_ast(*args):
            return ast.literal_eval(*args)

    start_time = time.time()

    logging.info(f"Using {THREAD_POOL} threads")

    data_path, metadata_path, report_path, run_id_path = file_handler(config)

    logging.getLogger().addHandler(
        logging.FileHandler(f"{metadata_path}/road_data_pipeline.log")
    )

    start_date = my_ast(config["user_settings"]["start_date"])
    end_date = my_ast(config["user_settings"]["end_date"])

    if not start_date and not end_date:

        # 2-month API data lag (date today minus two months)
        date_object_today = datetime.strptime(time.strftime("%Y-%m"), "%Y-%m")
        minus_two_months = date_object_today - relativedelta(months=2)

        year, month = map(int, minus_two_months.strftime("%Y %m").split())
        last_day_of_month = calendar.monthrange(year, month)[1]

        start_date = f"{year}-{month}-01"
        end_date = f"{year}-{month}-{last_day_of_month}"

    test_run = my_ast(config["user_settings"]["test_run"])
    generate_report = my_ast(config["user_settings"]["generate_report"])

    logging.info("Getting Road Sensor Lookup Table")

    sensor_tables, lookup_df = get_sites_by_sensor()
    lookup_df.to_csv(f"{str(metadata_path)}/road_data_sensor_lookup.csv", index=False)

    midas_metadata, tmu_metadata, tame_metadata = get_site_urls(
        sensor_tables, start_date, end_date
    )

    logging.info("Processed Road Sensor Lookup Table")

    if test_run:
        logging.info("Test Run")
        midas_metadata = midas_metadata[1:2]
        tmu_metadata = tmu_metadata[1:2]
        tame_metadata = tame_metadata[1:2]

    download(
        site_name="midas",
        start_date=start_date,
        end_date=end_date,
        metadata=midas_metadata,
        test_run=test_run,
        run_id_path=data_path,
    )

    download(
        site_name="tmu",
        start_date=start_date,
        end_date=end_date,
        metadata=tmu_metadata,
        test_run=test_run,
        run_id_path=data_path,
    )

    download(
        site_name="tame",
        start_date=start_date,
        end_date=end_date,
        metadata=tame_metadata,
        test_run=test_run,
        run_id_path=data_path,
    )

    if generate_report:
        run_reports(lookup_df, report_path, start_date, end_date)

    if api_run:
        dump_config(config, metadata_path, api_run=True)
    else:
        dump_config(config, metadata_path, api_run=False)

    gcp_storage = my_ast(config["user_settings"]["gcp_storage"])

    if gcp_storage:

        gcp_bucket_name = my_ast(config["user_settings"]["gcp_bucket_name"])
        gcp_folder_name = my_ast(config["user_settings"]["gcp_blob_name"])
        gcp_credentials = my_ast(config["user_settings"]["gcp_credentials"])

        gcp_upload_from_directory(
            run_id_path, gcp_bucket_name, gcp_folder_name, gcp_credentials
        )

    rm_dir = my_ast(config["user_settings"]["rm_dir"])

    if rm_dir:
        logging.info(
            f"Removing {run_id_path[run_id_path.find('output_data/'):].split('/')[1]} folder."
        )
        shutil.rmtree(Path(run_id_path))

    logging.info(f"Script Run Time: {(time.time() - start_time)/60} minutes.")


if __name__ == "__main__":

    config = configparser.ConfigParser()
    config.read("./config.ini")

    run(config)



import ast
import datetime
import glob
import logging
import os
from pathlib import Path

from google.cloud import storage


def file_handler(config):

    run_id = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S.%f")

    user_output_path = ast.literal_eval(config["user_settings"]["output_path"])

    if not user_output_path:
        raise ValueError("Please provide a valid output directory.")

    run_id_path_data = f"{user_output_path}/output_data/{run_id}/data/"
    run_id_path_metadata = f"{user_output_path}/output_data/{run_id}/metadata/"
    run_id_path_report = f"{user_output_path}/output_data/{run_id}/report/"
    run_id_path = f"{user_output_path}/output_data/{run_id}/"

    data_path = Path(run_id_path_data)
    metadata_path = Path(run_id_path_metadata)
    report_path = Path(run_id_path_report)

    logging.info(f"Making Data Directory at: {data_path}")
    data_path.mkdir(parents=True, exist_ok=True)

    logging.info(f"Making Metadata Directory at: {metadata_path}")
    metadata_path.mkdir(parents=True, exist_ok=True)

    logging.info(f"Making Report Directory at: {report_path}")
    report_path.mkdir(parents=True, exist_ok=True)

    return data_path, metadata_path, report_path, run_id_path


def dump_config(config, full_path, api_run):

    logging.info(f"Dumping config.ini for Run at {full_path}")

    if api_run:
        config_dict = config
    else:
        config_dict = {
            section: dict(config.items(section)) for section in config.sections()
        }

    with open(f"{str(full_path)}/config_metadata.txt", "w") as file:
        print(config_dict, file=file)


def gcp_upload_from_directory(
    directory_path: str,
    destination_bucket_name: str,
    destination_blob_name: str,
    gcp_credentials: str,
):

    os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = gcp_credentials
    storage_client = storage.Client()
    bucket = storage_client.get_bucket(destination_bucket_name)
    local_paths = glob.glob(directory_path + "/**", recursive=True)

    for local_file in local_paths:

        local_file = Path(local_file)
        remote_path = f"{destination_blob_name}/{str(local_file)[str(local_file).find('output_data/'):]}"

        if local_file.is_file():
            blob = bucket.blob(remote_path)
            blob.upload_from_filename(local_file)
            logging.info(
                f"Uploading {local_file} to Google Cloud Bucket: {destination_bucket_name} \n"
                f"Subfolder {destination_blob_name}"
            )



import glob
import logging
import warnings
from pathlib import Path

with warnings.catch_warnings():
    warnings.simplefilter(action="ignore", category=FutureWarning)
    import papermill as pm

from nbconvert.exporters import HTMLExporter
from nbconvert.preprocessors import TagRemovePreprocessor
from nbconvert.writers import FilesWriter
from traitlets.config import Config

warnings.simplefilter(action="ignore")
c = Config()
c.TemplateExporter.exclude_input = True
c.TagRemovePreprocessor.remove_cell_tags = ("remove_cell",)
c.TagRemovePreprocessor.remove_all_outputs_tags = ("remove_output",)
c.TagRemovePreprocessor.remove_input_tags = ("remove_input",)
c.TagRemovePreprocessor.enabled = True
c.HTMLExporter.preprocessors = ["nbconvert.preprocessors.TagRemovePreprocessor"]

exporter = HTMLExporter(config=c)
exporter.register_preprocessor(TagRemovePreprocessor(config=c), True)


def run_reports(data, full_path, start_date, end_date):

    logging.info(f"Generating HTML Report at {full_path}")

    input_path = "./report/road_data_report_template.ipynb"
    output_path = f"{full_path}/road_data_report.ipynb"

    report_date = f"{start_date} to {end_date}"
    report_title = f"__Road Data Sensors API Scraping Report__\n Date: {report_date}"

    params = {"title": report_title, "data": data.to_json()}

    logging.disable(logging.INFO)
    pm.execute_notebook(
        input_path=input_path,
        output_path=output_path,
        parameters=params,
        report_mode=True,
    )

    notebook_files = glob.glob(f"{str(full_path)}/*.ipynb")
    notebook_files = Path(notebook_files[0])

    (body, resources) = HTMLExporter(config=c).from_filename(notebook_files)
    fw = FilesWriter(config=c)
    fw.write(body, resources, notebook_name="road_data_report")
    logging.disable(logging.NOTSET)



import datetime

import pandas as pd
import requests


def get_site_urls(sensor_tables, start_date, end_date):

    start_date = datetime.datetime.strptime(start_date, "%Y-%m-%d").strftime("%d%m%Y")
    end_date = datetime.datetime.strptime(end_date, "%Y-%m-%d").strftime("%d%m%Y")

    midas_id = list(sensor_tables["midas"]["id"])
    midas_urls = [
        f"https://webtris.highwaysengland.co.uk/api/v1/reports/{start_date}/to/{end_date}/daily?sites={site}&page=1&page_size=40000"
        for site in midas_id
    ]

    midas_direction = sensor_tables["midas"]["direction"]
    midas_longitude = sensor_tables["midas"]["longitude"]
    midas_latitude = sensor_tables["midas"]["latitude"]
    midas_status = sensor_tables["midas"]["status"]
    midas_easting = sensor_tables["midas"]["easting"]
    midas_northing = sensor_tables["midas"]["northing"]

    midas_metadata = list(
        zip(
            midas_urls,
            midas_direction,
            midas_longitude,
            midas_latitude,
            midas_status,
            midas_easting,
            midas_northing,
        )
    )

    tmu_id = list(sensor_tables["tmu"]["id"])
    tmu_urls = [
        f"https://webtris.highwaysengland.co.uk/api/v1/reports/{start_date}/to/{end_date}/daily?sites={site}&page=1&page_size=40000"
        for site in tmu_id
    ]

    tmu_direction = sensor_tables["tmu"]["direction"]
    tmu_longitude = sensor_tables["tmu"]["longitude"]
    tmu_latitude = sensor_tables["tmu"]["latitude"]
    tmu_status = sensor_tables["tmu"]["status"]
    tmu_easting = sensor_tables["tmu"]["easting"]
    tmu_northing = sensor_tables["tmu"]["northing"]

    tmu_metadata = list(
        zip(
            tmu_urls,
            tmu_direction,
            tmu_longitude,
            tmu_latitude,
            tmu_status,
            tmu_easting,
            tmu_northing,
        )
    )

    tame_id = list(sensor_tables["tame"]["id"])
    tame_urls = [
        f"https://webtris.highwaysengland.co.uk/api/v1/reports/{start_date}/to/{end_date}/daily?sites={site}&page=1&page_size=40000"
        for site in tame_id
    ]

    tame_direction = sensor_tables["tame"]["direction"]
    tame_longitude = sensor_tables["tame"]["longitude"]
    tame_latitude = sensor_tables["tame"]["latitude"]
    tame_status = sensor_tables["tame"]["status"]
    tame_easting = sensor_tables["tame"]["easting"]
    tame_northing = sensor_tables["tame"]["northing"]

    tame_metadata = list(
        zip(
            tame_urls,
            tame_direction,
            tame_longitude,
            tame_latitude,
            tame_status,
            tame_easting,
            tame_northing,
        )
    )

    return midas_metadata, tmu_metadata, tame_metadata


def status_string_cleaner(x):
    if "eastbound" in str(x):
        return "eastbound"
    elif "northbound" in str(x):
        return "northbound"
    elif "southbound" in str(x):
        return "southbound"
    elif "westbound" in str(x):
        return "westbound"
    elif "clockwise" in str(x):
        return "clockwise"
    elif "anti-clockwise" in str(x):
        return "clockwise"
    elif "legacy site" in str(x):
        return "legacy site"
    elif "on connector" in str(x):
        return "carriageway connector"
    else:
        return x


def name_string_cleaner(x):
    if "MIDAS" in str(x):
        return "MIDAS"
    elif "TMU" in str(x):
        return "TMU"
    elif "TAME" in str(x):
        return "TAME"
    elif "Legacy Site" in str(x):
        return "Legacy Site"
    else:
        return x


def get_sites_by_sensor():

    url = "https://webtris.highwaysengland.co.uk/api/v1/sites"
    response = requests.get(url)

    lookup_df = pd.DataFrame.from_dict(response.json()["sites"])
    lookup_df.columns = [col.lower() for col in lookup_df.columns]
    lookup_df["id"] = lookup_df["id"].astype(int)

    lookup_df["direction"] = lookup_df["name"].str.split("; ").str[-1]
    lookup_df[["easting", "northing"]] = (
        lookup_df["name"].str.extract(r"(\d+;\d+)")[0].str.split(";", expand=True)
    )

    lookup_df["direction"] = lookup_df["direction"].str.lower()
    lookup_df["direction"] = lookup_df["direction"].apply(status_string_cleaner)

    lookup_df["name"] = lookup_df["name"].apply(name_string_cleaner)

    midas_df = lookup_df.query("name.str.contains('MIDAS', case = True)")
    tmu_df = lookup_df.query("name.str.contains('TMU', case = True)")
    tame_df = lookup_df.query("name.str.contains('TAME', case = True)")
    other_df = lookup_df.query(
        "name.str.contains('MIDAS|TAME|TMU', case = True)==False"
    )

    sensor_tables = {
        "midas": midas_df,
        "tmu": tmu_df,
        "tame": tame_df,
        "other": other_df,
    }

    return sensor_tables, lookup_df



import os
from pathlib import Path

from pkg_resources import get_distribution

__version__ = get_distribution("road_data_scraper").version

package_path = Path(__file__).parent
os.chdir(package_path)
